% :- type tabla           ---> a(meret, aknaszam, list(ref_pont)).
% :- type ref_pont        ---> p(sor, oszlop, ertek).
% :- type sor               == integer.
% :- type oszlop            == integer.
% :- type ertek             == integer.
% :- type aknaszam          == integer.
% :- type meret           ---> sor-oszlop.
% :- type pont            ---> sor-oszlop.
% :- type aknamezo          == list(pont).

% :- type file              ==  atom.
 
% :- pred aknakereso(tabla::in, aknamezo::out).


:- module(aknakereso_keret,
	  [
	   helyes_feladvany/1,  % helyes_feladvany(tabla::in)
	   aknakereso_be/2,	% aknakereso_be(file::in, tabla::out)
	   aknakereso_ki/3,	% aknakereso_ki(file::in, tabla::in,
	                        %         aknamezo::in)
	   megold/2,		% megold(file::in, file::in)
	   stopper/2		% stopper(file::in, file::in)
	  ]
	 ).

:- use_module(library(lists), [member/2,append/3]).

%-------------- Bejaratok ------------------------------------


% aknakereso_be(+File, -Feladvany):
% A File-bol beolvashato a Tabla altal megadott feladat.
aknakereso_be(File, Feladvany) :-
	open_file(File, read, Stream),
	call_cleanup(read_in(Stream, Characters),
		     close(Stream)),
	parse_data(Characters, Feladvany).

% aknakereso_ki(+File, +Feladvany, +Aknamezo):
% A Feladvany feladat Aknamezo megoldasat kiirja a File allomanyba.
aknakereso_ki(File, Feladvany, Aknamezo) :-
	open_file(File, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(pretty_print(Feladvany, Aknamezo),
		     (set_output(OOut),close(Stream))).

% megold(+FileIn, +FileOut):
% A FileIn allomanybol beolvasott feladat minden megoldasat
% kiirja a FileOut allomanyba.
megold(FileIn, FileOut) :-
	aknakereso_be(FileIn, Tabla),
	ensure_loaded(user:aknakereso),
	open_file(FileOut, write, Stream),
	current_output(OOut),
	set_output(Stream),
	call_cleanup(print_solutions(Tabla),
		     (set_output(OOut),close(Stream))).

% stopper(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper(FileIn, FileOut) :-
	ensure_loaded(user:aknakereso),
	statistics(runtime, [T0|_]),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	format('Feladat: ~a, futasi ido ~3d s~n', [FileIn,T]).

helyes_feladvany(a(Meret,Aknaszam,RefPL)) :-
	% a palya merete helyes:
	helyes_pont(Meret,inf-inf),
        % nincs nem helyes eleme a RefPL listanak:
 	\+ (   member(Ref, RefPL),  
	       \+ helyes_referencia(Ref, Meret) 
	   ),
	% a RefPL lista lexikografikusan rendezett:
 	sort(RefPL, RefPL),
	% A RefPL lista minden eleme kulonbozo (Sor,Oszlop) kordinatara
	% vonatkozik (a rendezettseg miatt elegendo ezt a szomszedos
	% elemekre kikotni):
	\+ append(_, [i(S,O,_),i(S,O,_)|_], RefPL),
	% Eleg szabad mezo van az aknak szamara:
	Meret = Sor-Oszlop,
	Terulet is Sor * Oszlop,
	length(RefPL, Szammezok),
	Terulet - Szammezok >= Aknaszam.


%-------------- Ellenorzesek ---------------------------------

helyes_pont(S-O, MaxS-MaxO) :-
	integer(S), S >=1, S =< MaxS,
	integer(O), O >=1, O =< MaxO.

helyes_referencia(p(Sor,Oszlop,Ertek), Meret) :-
	helyes_pont(Sor-Oszlop, Meret),
	0 =< Ertek, Ertek =< 8.


%-------------- Allomanyok kezelese --------------------------

% open_file(+File, +Mode, -Stream):
% A File allomanyt megnyitja a Stream folyam formajaban Mode modban.
open_file(user, M, S) :-
	!, standard_stream(M, S).
open_file(File, Mode, Stream) :-
	open(File, Mode, Stream).

% standard_stream(?Mode, ?Stream):
% A Mode modu szabvanyos folyam a Stream.
standard_stream(read, user_input).
standard_stream(write, user_output).

%-------------- Beolvasas ------------------------------------

% read_in(+S, -L):
% S folyambol beolvashato karakterek listaja L.
read_in(S, L) :-
	get_code(S, C),
	read_in(C, S, L).

% read_in(+C, +S, -L):
% Ha C az S folyambol elozoleg beolvasott karakter, akkor L az S-bol
% beolvashato maradek karakterek listaja, C-t is beleertve.
% Ha C = -1, akkor L = [].
read_in(C, _, L) :-
	C =:= -1, !, L = [].
read_in(C, S, [C|L]) :-
	get_code(S, C1),
	read_in(C1, S, L).

% parse_data(+Chars, -Table):
% Chars fuzerbol kielemezheto egy Feladvany feladvanyleiro
parse_data(Chars, Feladvany) :-
	descriptor(Feladvany, Chars, Rest),
	correct_syntax(Rest),
	helyes_feladvany(Feladvany), !.
parse_data(_, _) :-
	warning('A bemeneti allomany nem megfelelo formatumu.',[]),
	fail.

% correct_syntax(+Rest):
% A Rest fuzer a bemenet helyes maradeka, azaz ures.
correct_syntax(Rest) :-
	Rest = [_|_], !,
	chars_of_type(mid_line, L, Rest, _),
	warning('Nem vart karakter(ek): "~s"', [L]),
	fail.
correct_syntax(_).

%-------------- DCG elemzes ----------------------------------

% Kielemezheto egy feladvany-leiro
descriptor(a(R-C,I,RefPs)) -->
	nat_number(R), nat_number(C),
	newline,
	nat_number(I),
	newline,
	triplet_list(RefPs),
	chars_of_type(layout, _).

% Kielemezheto egy (tizes szamrendszerbeli) Int termeszetes szam, amit
% esetleg szokozok eloznek meg.
nat_number(Int) -->
	chars_of_type(horizontal, _),
	chars_of_type(digit, Ds),
	{Ds = [_|_], number_codes(Int, Ds)}.

% triplet_list(RefPs)
% Kielemezheto egy harmasokbol allo RefPs lista.
triplet_list([p(R,C,V)|RefPs]) -->
	nat_number(R), nat_number(C), nat_number(V),
	newline, !,
	triplet_list(RefPs).
triplet_list([]) --> [].

newline -->
	chars_of_type(horizontal, _),
	char_of_type(vertical, _).


% layout_char(Dir, C):
% horizontal_layout_char(C): C  egy vizszintes iranyu
% nem lathato karakter.
horizontal_layout_char(0' ).
horizontal_layout_char(0'\t).
horizontal_layout_char(0'\r).

% vertical_layout_char(C): C  egy fuggoleges iranyu
% nem lathato karakter.
vertical_layout_char(0'\n).
vertical_layout_char(0'\f).
vertical_layout_char(0'\v).

% Kielemezheto egy Type tipusu C karakter.
char_of_type(Type, C) -->
	[C], {char_type(Type, C)}.

% chars_of_type(Type, Cs) -->
% Kielemezheto Type tipusu karakterek egy Cs listaja.	
chars_of_type(Type, [C|Cs]) -->
	char_of_type(Type, C), !,
	chars_of_type(Type, Cs).
chars_of_type(_, []) --> [].

% char_type(Type, C) :
% A C karakter Type tipusu.
char_type(layout, C) :-
	(   vertical_layout_char(C) -> true
	;   horizontal_layout_char(C)
	).
char_type(digit, D) :-
	D >= 0'0, D =< 0'9.
char_type(mid_line, C) :-
	\+ vertical_layout_char(C).
char_type(horizontal, C) :-
	horizontal_layout_char(C).
char_type(vertical, C) :-
	vertical_layout_char(C).



%-------------- Kiiratas -------------------------------------

% print_solutions(+Descr):
% A Tabla altal megadott feladat osszes megoldasat
% kiirja az aktualis kimenetre.
print_solutions(Descr) :-
	user:aknakereso(Descr, Minefield),
	pretty_print(Descr, Minefield),
	fail.
print_solutions(_).

% pretty_print(Descr)
% Kiirja a Descr tablat az aknamezo nelkul.
pretty_print(Descr) :-
	pretty_print(Descr, []).

% pretty_print(Descr, Minefield)
% Kiirja a Descr tabla megoldasait Minefield aknamezo elem eseten
pretty_print(Descr, Minefield) :-
	Descr = a(R-C,_,_),
	print_separator(C),
	setof(Point:Value,
	      table_elem(Descr, Minefield, Point, Value),
	      Elems),
	print_lines(1, R, C, Elems).

% table_elem(Descr, Minefield, Point, Value):
% A Descr feladvany Minefield megoldasanak megfelelo tablan
% a Point koordinataju pontban megjelenitendo a Value ertek.
table_elem(a(_,_,RefPs), Minefield, Point, Value) :-
	(   member(p(R,C,Value), RefPs), Point = R-C
	;   member(Point, Minefield),
	    Value = '*'
	).

% print_lines(R0, R, C, Elems):
% Kiirja az R x C meretu tabla R0-adik sorat,
% az Elems tartalomlistanak megfeleloen.
% A tartalomlista (Sor-Oszlop):Ertek alaku elemekbol all,
% es azt irja elo, hogy a (Sor,Oszlop) pozicion megjelenitendo
% az Ertek konstans.
print_lines(R0, R, _, _) :-
	R0 > R, !.
print_lines(R0, R, C, Elems0) :-
	print_line(1, C, R0, Elems0, Elems1),
	print_separator(C),
	R1 is R0+1,
	print_lines(R1, R, C, Elems1).

% print_line(C0, C, R0, Elems0, Elems):
% Kiirja az R x C meretu tabla R0-adik soranak a C0 oszlopban
% kezdodo reszet az Elems0 tartalomlistanak megfeleloen.
% Az Elems argumentumban visszaadja a tartalomlistanak
% az R0 sor utani reszet.
print_line(C0, C, _, Elems, Elems) :-
	C0 > C, !,
	format('|~n', []).
print_line(C0, C, R0, Elems0, Elems) :-
	field(R0-C0, Value, Elems0, Elems1),
	format('| ~w ', [Value]),
	C1 is C0+1,
	print_line(C1, C, R0, Elems1, Elems).

% field(Point, Value, Elems0, Elems):
% A Point koordinataju pontban, az Elems0 tartalomlista szerint
% megjelenitendo ertek a Value. Elems a Point utani resznek megfelelo
% tartalomlista.
field(Point, Value, [Point:Value|Elems], Elems) :- !.
field(_, ' ', Elems, Elems).
	
% print_separator(N)
% Kiir egy N hossz elvalaszto sort.
print_separator(0) :-
	!, write('+'), nl.
print_separator(N) :-
	write('+---'),
	N1 is N-1,
	print_separator(N1).

%-------------- Seged-eljarasok ------------------------------

% warning(+Txt, +Args):
% Kiirja a Txt szoveggel es Args argumentumlistaval
% megadott figyelmezteto szoveget.
warning(Txt, Args) :-
	print_message(warning, format(Txt, Args)).

%-------------- ISO kompatibilitas ---------------------------

% Az alabbi cel az ISO Prolog szabvanynak meg nem megfelelo, de
% DEC10 kompatiblis Prolog rendszerekben valo futtatashoz szukseges.

:- (   predicate_property(get_code(_,_), _) -> true
   ;   assert((get_code(S, C) :- get0(S, C))),
       assert((put_code(C) :- put(C))),
       assert((number_codes(N, C) :- number_chars(N, C)))
   ).
