type sor      = int
type oszlop   = int
type ertek    = int
type aknaszam = int
type meret    = sor * oszlop
type pont     = sor * oszlop
type ref_pont = sor * oszlop * ertek
type tabla    = meret * aknaszam * ref_pont list
type aknamezo = pont list;

app load ["Binarymap","Int", "Listsort", "Time", "Timer"];
use "aknakereso.sml";

exception Akna of string;

(* helyes_feladvany tabla = Az aknakereso fggvny els paramtere
      mindenkppen kielgti ezt a prediktumot.
 *)
fun helyes_feladvany (Meret, Aknaszam, RefPL) =
    let
	val (MaxS, MaxO) = Meret
	val Terulet = MaxS * MaxO
	val RefPontSzam = length RefPL
	fun helyes_pont (Sor, Oszlop) =
	      1 <= Sor andalso 1 <= Oszlop andalso
	      Sor <= MaxS andalso Oszlop <= MaxO
	fun helyes_ref_pont (Sor, Oszlop, Ertek)  =
	      helyes_pont (Sor, Oszlop) andalso
	      0 <= Ertek andalso Ertek <= 8;

	fun lexikografikusan_hasonlit ((x1,y1,_), (x2,y2,_)) =
	      if      x1 < x2 orelse x1 = x2 andalso y1 < y2 then LESS
	      else if x1 > x2 orelse x1 = x2 andalso y1 > y2 then GREATER
	      else EQUAL
	fun szomszedok_kulonbozok (rp1::(rps as rp2::_)) =
	      lexikografikusan_hasonlit (rp1, rp2) <> EQUAL
	      andalso szomszedok_kulonbozok rps
	  | szomszedok_kulonbozok _ = true
    in 
	(* a tbla mretezse, az aknaszm megadsa helyes: *)
	1 <= MaxS andalso 1 <= MaxO andalso
	Terulet - RefPontSzam >= Aknaszam andalso
	(* a RefPL lista minden eleme helyes: *)
        (List.all (fn x => x) (map helyes_ref_pont RefPL)) andalso
	(* a RefPL lista lexikografikusan rendezett: *)
	Listsort.sorted lexikografikusan_hasonlit RefPL andalso
	(* A RefPL lista minden eleme klnbz (Sor,Oszlop)
	 koordintra vonatkozik (a rendezettsg miatt elegend
	 ezt a szomszdos elemekre kiktni): *)
	szomszedok_kulonbozok RefPL
    end


(* aknakereso_be f = Az f szvegfjlbl beolvasott tbla
   aknakereso_be : string -> tabla
 *)
fun aknakereso_be f =
    let
	fun par [a,b] = (a,b)
	  | par _ =
	    raise Akna("Szintaktikusan hibs mretmegads a(z) " ^ f ^ " fjlban.")

	fun harmas [a,b,c] = (a,b,c)
	  | harmas _ =
	    raise Akna("Szintaktikusan hibs mezler a(z) " ^ f ^ " fjlban.")

	fun fuzerbol l2n = l2n o
	                   (List.mapPartial Int.fromString) o
			   (String.tokens Char.isSpace)

        (* az MOSML-ben s az Alice-ben ms a tpusuk *)
        (* Alice-vltozatok
	  fun inputLine s = Option.valOf(TextIO.inputLine s);
          fun sort compare xs = List.sort compare xs;
	*)
        (* SML-vltozatok *)
	  fun inputLine s = TextIO.inputLine s;
          fun sort compare xs = Listsort.sort compare xs;
        (* vltozatok vge *)

	fun refpontok_be s =
	    if TextIO.endOfStream s then []
	    else (fuzerbol harmas) (inputLine s) :: refpontok_be s

	fun compare ((y1,x1,_), (y2,x2,_)) =
	    if      y1 > y2 then GREATER
	    else if y1 < y2 then LESS
	    else if x1 > x2 then GREATER
	    else if x1 < x2 then LESS
	    else raise Akna("Ismtld rcspont a(z) " ^ f ^ " fjlban.")

	val istr      = TextIO.openIn f
	val meret     = (fuzerbol par) (inputLine istr)
	val aknaszam  = Option.valOf(Int.fromString(inputLine istr))
	val refpontok = sort compare (refpontok_be istr)
	val tabla     = (meret, aknaszam, refpontok) before TextIO.closeIn istr
    in
	if helyes_feladvany tabla then tabla
	else raise Akna("Hibs tbla a(z) " ^ f ^ " fjlban.")
    end;

(* aknarajz (tabla, aknamezo) = az aknamezonek megfelel aknarajz tabla-n
   aknarajz : tabla * aknamezo -> string
 *)
fun aknarajz (((m,sz), _, rpk), aknak) =
    let
	fun paircmp ((a,b), (c,d)) =
	    case Int.compare(a,c) of
		EQUAL => Int.compare(b,d)
	      | other => other

	fun insert ((S,O,c), bmap) = Binarymap.insert(bmap, (S,O), c)

	val pk = Binarymap.mkDict paircmp
	val pk = foldl insert pk (map (fn (S,O) => (S,O,#"*")) aknak)
	val pk = foldl insert pk (map (fn (S,O,v) => (S,O,chr(ord #"0" + v))) rpk)

	val tab   = String.concat o List.tabulate
	val vonal = tab (sz, fn _ => "+---") ^ "+\n"

	fun cellarajz S O = case Binarymap.peek(pk, (S+1,O+1)) of
	                        SOME c => " " ^ str c ^ " |"
			      | NONE   => "   |"

	fun sorrajz i = "|" ^ tab (sz, cellarajz i) ^ "\n" ^ vonal
    in
	vonal ^ tab (m, sorrajz)
    end

(* aknarajzok (t, mezok) = a mezok megoldslistnak megfelel rajzok a
      t tbln, res sorokkal elvlasztva
   aknarajzok : tabla * aknamezo list -> string
 *)
fun aknarajzok (t, mezok) =
    foldl (fn (aknak,s) => s ^ aknarajz(t, aknak)) "" mezok

(* aknakereso_ki(m, t, mezok) = A t tbla mezok megoldslistjt olvashat
      formban kirja az m szveges llomnyba.
   aknakereso_ki : string * tabla * aknamezo list -> unit
 *)
fun aknakereso_ki(m, t, mezok) =
      let val ostr = TextIO.openOut m
	  val _ = TextIO.output(ostr, aknarajzok(t, mezok))
      in
	 TextIO.closeOut ostr
      end
  
(* megold(f, m) = Beolvas egy feladvnyt az f szveges llomnybl s
      kirja az sszes megoldst az m szveges llomnyba. Ehhez
      termszetesen szksge van az aknakereso fggvnyre. Eredmnye az f
      nevet, a megoldsok szmt s a futsi idt mp-ben tartalmaz
      fzr.
   megold : string * string -> string
 *)
fun megold (f, m) =
    let val t = aknakereso_be f
	val start = Timer.startCPUTimer()
	val mezok = aknakereso t
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = aknakereso_ki(m, t, mezok)
    in
	"Feladvny: " ^ f ^ ", " ^
	Int.toString(length mezok) ^ " megolds, " ^
	"id: " ^ Time.fmt 2 tim ^ " mp\n"
    end
